module PaperCropper
  module ModelExtension
    # Class methods
    # #########################################################################
    # Initializes attachment cropping in your model
    #
    #   crop_attached_file :avatar
    #
    # You can also define an initial aspect ratio for the crop and preview box through opts[:aspect]
    #
    #   crop_attached_file :avatar, :aspect => "4:3"
    #
    # Or unlock it
    #
    #   crop_attached_file :avatar, :aspect => false
    #
    # @param attachment_name [Symbol] Name of the desired attachment to crop
    # @param opts [Hash]
    # @option opts [Range, String, FalseClass] :aspect
    module ClassMethods
      def crop_attached_file(attachment_name, opts = {})
        opts = opts.dup

        include PaperCropper::ModelExtension::InstanceMethods

        aspect = normalize_aspect opts[:aspect]
        send :define_method, :"#{attachment_name}_aspect" do
          aspect.first.to_f / aspect.last.to_f if aspect
        end

        if respond_to? :attachment_definitions
          # for Paperclip <= 3.4
          definitions = attachment_definitions
        else
          # for Paperclip >= 3.5
          definitions = Paperclip::Tasks::Attachments.instance.definitions_for(self)
        end
        processors = definitions[attachment_name][:processors] ||= []
        unless processors.include? :paper_cropper
          processors << :paper_cropper
        end

        after_save :"reprocess_to_crop_#{attachment_name}_attachment"
        after_update {
          if send(:attachment_changed?, attachment_name)
            self.update_column(:"#{attachment_name}_updated_at", Time.now)
          end
        }
      end

      # Returns a valid and normalized value for aspect ratio
      # It will return 1.. if aspect is nil or a invalid string
      # @param aspect [Range, String, FalseClass]
      #
      # @return [Range]
      def normalize_aspect(aspect)
        if aspect.kind_of?(String) && aspect =~ PaperCropper::RegExp::ASPECT
          Range.new *aspect.split(':').map(&:to_i)
        elsif aspect.kind_of?(Range)
          return aspect.first.to_i..aspect.last.to_i
        elsif aspect == false
          return aspect
        else
          1..1
        end
      end
    end

    # Instance methods
    # #########################################################################
    module InstanceMethods
      # Asks if the attachment received a crop process
      # @param  attachment_name [Symbol]
      #
      # @return [Boolean]
      def cropping?(attachment_name)
        !self.send(:"#{attachment_name}_crop_x").blank? &&
          !self.send(:"#{attachment_name}_crop_y").blank? &&
          !self.send(:"#{attachment_name}_crop_width").blank? &&
          !self.send(:"#{attachment_name}_crop_height").blank?
      end

      def attachment_changed?(attachment_name)
        ["#{attachment_name}_name", "#{attachment_name}_content_type",
          "#{attachment_name}_file_size", "#{attachment_name}_updated_at",
          "#{attachment_name}_crop_x", "#{attachment_name}_crop_y",
          "#{attachment_name}_crop_width", "#{attachment_name}_crop_height",
          "#{attachment_name}_crop_ratio"
        ].any? { |attr|
          method= "#{attr}_changed?".to_sym
          if respond_to?(method)
            self.send(method)
          end
        }
      end

      # Uses method missing to responding the model callback for reprocess the image
      def method_missing(method, *args)
        if method.to_s =~ PaperCropper::RegExp::CALLBACK
          reprocess_cropped_attachment(
            method.to_s.scan(PaperCropper::RegExp::CALLBACK).flatten.first.to_sym
          )
        else
          super
        end
      end

      private
      # Saves the attachment if the crop attributes are present
      # @param  attachment_name [Symbol]
      def reprocess_cropped_attachment(attachment_name)
        if cropping?(attachment_name)
          attachment_instance = send(attachment_name)
          attachment_instance.assign(attachment_instance)
          attachment_instance.save
        end
      end
    end
  end
end


# ActiveRercord support
# #########################################################################
if defined? ActiveRecord::Base
  ActiveRecord::Base.class_eval do
    extend  PaperCropper::ModelExtension::ClassMethods
  end
end
